############################################################
# Program is part of PyAPS                                 #
# Copyright 2012, by the California Institute of Technology#
# Contact: earthdef@gps.caltech.edu                        #
# Modified by A. Benoit and R. Jolivet 2019                #
# Ecole Normale Superieure, Paris                          #
# Contact: insar@geologie.ens.fr                           #
############################################################
# Add more utils functions, Zhang Yunjun, April 2019
#     get_lat_lon()
#     read_(meta)data()
#     read_isce_xml/data/lalo_ref()
#     read_roipac_data/rsc()


import os
import sys
import time
import numpy as np



def snwe2str(snwe):
    """Get area extent in string"""
    if not snwe:
        return None
    s, n, w, e = snwe

    area = ''
    area += '_S{}'.format(abs(s)) if s < 0 else '_N{}'.format(abs(s))
    area += '_S{}'.format(abs(n)) if n < 0 else '_N{}'.format(abs(n))
    area += '_W{}'.format(abs(w)) if w < 0 else '_E{}'.format(abs(w))
    area += '_W{}'.format(abs(e)) if e < 0 else '_E{}'.format(abs(e))

    return area



############### Read ISCE / ROIPAC file ###############

def get_lat_lon(metafile):
    """Get 2D lat and lon from rsc/xml file
    Args:
        * metafile (str or dict) : path to metadata file, or dict of metadata
    """
    if isinstance(metafile, str):
        fext = os.path.splitext(metafile)[1]
        if fext == '.rsc':
            meta = read_roipac_rsc(metafile)
        elif fext == '.xml':
            meta = read_isce_xml(metafile)
        else:
            raise ValueError('unrecognized file extension: {}'.format(fext))
    elif isinstance(metafile, dict):
        meta = {}
        for key, value in metafile.items():
            meta[key] = value
    length, width = int(meta['FILE_LENGTH']), int(meta['WIDTH'])

    # get bbox
    if 'Y_FIRST' in meta.keys():
        # geo coordinates
        lat0 = float(meta['Y_FIRST'])
        lon0 = float(meta['X_FIRST'])
        lat_step = float(meta['Y_STEP'])
        lon_step = float(meta['X_STEP'])
        lat1 = lat0 + lat_step * (length - 1)
        lon1 = lon0 + lon_step * (width - 1)
    else:
        # radar coordinates
        lats = [float(meta['LAT_REF{}'.format(i)]) for i in [1,2,3,4]]
        lons = [float(meta['LON_REF{}'.format(i)]) for i in [1,2,3,4]]
        lat0 = np.mean(lats[0:2])
        lat1 = np.mean(lats[2:4])
        lon0 = np.mean(lons[0:3:2])
        lon1 = np.mean(lons[1:4:2])

    # bbox --> 2D array
    lat, lon = np.mgrid[lat0:lat1:length*1j,
                        lon0:lon1:width*1j]
    return lat, lon


def read_data(fname, dname='inc'):
    """Read 2D data from ISCE / ROIPAC file"""
    # get meta file extension
    meta_exts = [i for i in ['.xml', '.rsc'] if os.path.isfile(fname+i)]
    if len(meta_exts) == 0:
        raise FileNotFoundError('No metadata file found for data file: {}'.format(fname))
    meta_ext = meta_exts[0]

    if meta_ext == '.xml':
        data = read_isce_data(fname, dname=dname)
    elif meta_ext == '.rsc':
        data = read_roipac_data(fname)
    return data


def read_metadata(fname, latfile=None, lonfile=None, full=False, verbose=False):
    '''Reading metadata for ISCE or ROIPAC style data file

    Args:
        * fname (str): Path to the ROIPAC or ISCE data file.

    Returns:
        * lat  (np.ndarray): Array of lat of the 4 corners.
        * lon  (np.ndarray): Array of lon of the 4 corners.
        * nx   (int)       : Number of range bins.
        * ny   (int)       : Number of azimuth lines.
        * dpix (float)     : Average pixel spacing.
        * meta (dict)      : Dictionary of metadata [return if full==True]
    '''
    # get meta file extension
    meta_exts = [i for i in ['.rsc','.xml'] if os.path.isfile(fname+i)]
    if len(meta_exts) == 0:
        raise FileNotFoundError('No metadata file found for data file: {}'.format(fname))
    meta_ext = meta_exts[0]

    if verbose:
        print("PROGRESS: READING {} FILE".format(fname+meta_ext))

    # read metadata
    meta = {}
    if meta_ext == '.rsc':
        meta = read_roipac_rsc(fname+meta_ext)

    elif meta_ext == '.xml':
        meta = read_isce_xml(fname+meta_ext)

        # default latfile
        if not latfile:
            latfile = os.path.join(os.path.dirname(fname),'lat.rdr')
            if not os.path.isfile(latfile):
                raise FileNotFoundError("No latitude file found in ISCE style!")

        # default lonfile
        if not lonfile:
            lonfile = os.path.join(os.path.dirname(fname),'lon.rdr')
            if not os.path.isfile(lonfile):
                raise FileNotFoundError("No longitude file found in ISCE style!")

        meta.update(get_isce_lalo_ref(latfile, lonfile))

    # prepare output
    nx = int(meta['WIDTH'])
    ny = int(meta['FILE_LENGTH'])
    lat = np.zeros((4,1))
    lon = np.zeros((4,1))
    lat[0] = float(meta['LAT_REF1'])
    lon[0] = float(meta['LON_REF1'])
    lat[1] = float(meta['LAT_REF2'])
    lon[1] = float(meta['LON_REF2'])
    lat[2] = float(meta['LAT_REF3'])
    lon[2] = float(meta['LON_REF3'])
    lat[3] = float(meta['LAT_REF4'])
    lon[3] = float(meta['LON_REF4'])
    rgpix = 90.0
    azpix = 90.0
    dpix = np.sqrt(6.25*rgpix*rgpix + azpix*azpix)
    if full:
        return lon,lat,nx,ny,dpix,meta
    else:
        return lon,lat,nx,ny,dpix


def read_isce_xml(xmlfile):
    """Read ISCE XML file into dict.
    Add from mintpy.utils.readfile.py by Zhang Yunjun
    """
    import xml.etree.ElementTree as ET
    meta = {}
    root = ET.parse(xmlfile).getroot()
    if root.tag.startswith('image'):
        for child in root.findall('property'):
            key = child.get('name')
            value = child.find('value').text
            meta[key] = value

        # Read lat/lon info for geocoded file
        # in form: root/component coordinate*/property name/value
        for coord_name, prefix in zip(['coordinate1', 'coordinate2'], ['X', 'Y']):
            child = root.find("./component[@name='{}']".format(coord_name))
            if ET.iselement(child):
                v_step  = float(child.find("./property[@name='delta']").find('value').text)
                v_first = float(child.find("./property[@name='startingvalue']").find('value').text)
                if abs(v_step) < 1. and abs(v_step) > 1e-7:
                    meta['{}_STEP'.format(prefix)] = v_step
                    meta['{}_FIRST'.format(prefix)] = v_first - v_step / 2.

    # convert key name from isce to roipac
    isce2roipacKeyDict = {
        'width':'WIDTH',
        'length':'FILE_LENGTH',
    }
    for key,value in isce2roipacKeyDict.items():
        meta[value] = meta[key]
    return meta


def read_isce_data(fname, dname=None):
    """Read ISCE data file"""
    # read xml file
    xml_file = fname+'.xml'
    meta = read_isce_xml(xml_file)

    # get data_type
    dataTypeDict = {
        'byte': 'bool_',
        'float': 'float32',
        'double': 'float64',
        'cfloat': 'complex64',
    }
    data_type = dataTypeDict[meta['data_type'].lower()]

    width = int(meta['width'])
    length = int(meta['length'])
    num_band = int(meta['number_bands'])
    band = 1
    if fname.startswith('los') and dname and dname.startswith('az'):
        band = 2

    # read
    data = np.fromfile(fname, dtype=data_type, count=length*width*num_band).reshape(-1, width*num_band)
    data = data[:, width*(band-1):width*band]
    return data


def get_isce_lalo_ref(lat_file, lon_file):
    """Get LAT/LON_REF1/2/3/4 value from ISCE lat/lon.rdr file
    Add from mintpy.prep_isce.py by Zhang Yunjun
    """

    def get_nonzero_row_number(data, buffer=2):
        """Find the first and last row number of rows without zero value
        for multiple swaths data
        """
        if np.all(data):
            r0, r1 = 0 + buffer, -1 - buffer
        else:
            row_flag = np.sum(data != 0., axis=1) == data.shape[1]
            row_idx = np.where(row_flag)[0]
            r0, r1 = row_idx[0] + buffer, row_idx[-1] - buffer
        return r0, r1

    meta = {}
    # read LAT/LON_REF1/2/3/4 from lat/lonfile
    lat = read_isce_data(lat_file)
    r0, r1 = get_nonzero_row_number(lat)
    meta['LAT_REF1'] = str(lat[r0, 0])
    meta['LAT_REF2'] = str(lat[r0, -1])
    meta['LAT_REF3'] = str(lat[r1, 0])
    meta['LAT_REF4'] = str(lat[r1, -1])

    lon = read_isce_data(lon_file)
    r0, r1 = get_nonzero_row_number(lon)
    meta['LON_REF1'] = str(lon[r0, 0])
    meta['LON_REF2'] = str(lon[r0, -1])
    meta['LON_REF3'] = str(lon[r1, 0])
    meta['LON_REF4'] = str(lon[r1, -1])
    return meta


def read_roipac_rsc(rscfile):
    """Read ROIPAC style RSC file into dict"""
    meta = {}
    f = open(rscfile,'r')
    line = f.readline()
    while line:
        llist = line.split()
        if len(llist)>0 :
            meta[llist[0]] = llist[1]
        line = f.readline()
    f.close()
    return meta


def read_roipac_data(fname):
    """Read ROIPAC data file"""
    # get length / width
    rsc_file = fname+'.rsc'
    meta = read_roipac_rsc(rsc_file)
    length = int(meta['FILE_LENGTH'])
    width = int(meta['WIDTH'])

    # get datatype / band
    data_type = 'float32'
    num_band = 1
    band = 1
    fext = os.path.splitext(fname)[1]
    if fext == '.dem':
        data_type = 'int16'
    elif fext == '.hgt':
        num_band = 2
        band = 2

    # read
    data = np.fromfile(fname, dtype=data_type, count=length*width*num_band).reshape(-1, width*num_band)
    data = data[:, width*(band-1):width*band]
    return data

############### Read ISCE / ROIPAC file - end #########



############### obsolete: RSC for radar ###############

def rd_rsc(inname,full=False,verbose=False):
    '''Reading a ROI-PAC style RSC file.

    Args:
        * inname (str): Path to the RSC file.

    Returns:
        * lat  (np.ndarray): Array of lat of the 4 corners.
        * lon  (np.ndarray): Array of lon of the 4 corners.
        * nx   (int)       : Number of range bins.
        * ny   (int)       : Number of azimuth lines.
        * dpix (float)     : Average pixel spacing.

    .. note:: 
        Currently set up to work with SIM_nrlks.hgt from ROI-PAC.'''

    if verbose:
        print("PROGRESS: READING %s RSC FILE" %inname)

    rpacdict = {}
    infile = open(inname+'.rsc','r')
    line = infile.readline()
    rpacdict['LAT_REF1']=0.0
    rpacdict['LON_REF1']=0.0
    rpacdict['LAT_REF2']=0.0
    rpacdict['LON_REF2']=0.0
    rpacdict['LAT_REF3']=0.0
    rpacdict['LON_REF3']=0.0
    rpacdict['LAT_REF4']=0.0
    rpacdict['LON_REF4']=0.0
    while line:
        llist = line.split()
        if len(llist)>0 :
            rpacdict[llist[0]] = llist[1]
        line = infile.readline()
    infile.close()

    # prepare output
    nx = int(rpacdict['WIDTH'])
    ny = int(rpacdict['FILE_LENGTH'])
    lat=np.zeros((4,1))
    lon=np.zeros((4,1))
    lat[0] = float(rpacdict['LAT_REF1'])
    lon[0] = float(rpacdict['LON_REF1'])
    lat[1] = float(rpacdict['LAT_REF2'])
    lon[1] = float(rpacdict['LON_REF2'])
    lat[2] = float(rpacdict['LAT_REF3'])
    lon[2] = float(rpacdict['LON_REF3'])
    lat[3] = float(rpacdict['LAT_REF4'])
    lon[3] = float(rpacdict['LON_REF4'])
    rgpix = 90.0   # float(rpacdict['RANGE_PIXEL_SIZE'])
    azpix = 90.0   # float(rpacdict['AZIMUTH_PIXEL_SIZE'])
    dpix = np.sqrt(6.25*rgpix*rgpix+azpix*azpix)
    if full:
        return lon,lat,nx,ny,dpix,rpacdict
    else:
        return lon,lat,nx,ny,dpix
#######################Finished rd_rsc###########################


###############Reading input RSC file for geo###############
def geo_rsc(inname,full=False,verbose=False):
    '''Reading a ROI-PAC style geocoded rsc file.

    Args:
        * inname (str): Path to the RSC file.

    Returns:
        * lon (np.ndarray): Array of min and max lon values.
        * lat (np.ndarray): Array of min and max lat values.
        * nx  (int)       : Number of lon bins.
        * ny  (int)       : Number of lat bins.

    .. note:: 
        Currently set up to work with dem.rsc file from ROI-PAC.'''

    if verbose:
        print("PROGRESS: READING %s RSC FILE" %inname)

    rpacdict = {}
    infile = open(inname+'.rsc','r')
    line = infile.readline()
    while line:
        llist = line.split()
        if len(llist)>0 :
            rpacdict[llist[0]] = llist[1]
        line = infile.readline()
    infile.close()

    nx = int(rpacdict['WIDTH'])
    ny = int(rpacdict['FILE_LENGTH'])
    lat=np.zeros((2,1))
    lon=np.zeros((2,1))
    lat[1] = float(rpacdict['Y_FIRST'])
    lon[0] = float(rpacdict['X_FIRST'])
    if(lon[0] < 0):
        lon[0] = lon[0] + 360.0

    dx = float(rpacdict['X_STEP'])
    dy = float(rpacdict['Y_STEP'])

    lat[0] = lat[1] + dy*ny
    lon[1] = lon[0] + dx*nx
    
    if full:
        return lon,lat,nx,ny,rpacdict
    else:
        return lon,lat,nx,ny

############### obsolete: RSC for radar - end #########



############### obsolete: simple geo2radar ############

def lonlat2rdr(lon, lat, lonlist, latlist, plotflag=False):
    '''
    Transfer the lat lon coordinates of the weather stations into the index
    range and azimuth coordinates of the radar scene.

    Args:
        * lon               : Longitude array of the radar scene size=(ny,nx)
        * lat               : Latitude array of the radar scene size=(ny,nx)
        * lonlist           : Longitude list of the weeather stations
        * latlist           : Latitude list of the weather stations.

    Kwargs:
        * plotflag          : Plot something to check. (default is False)
    
    note:
        Mapping function is :
        * range = a1*lat+b1*lon+c1
        * azimu = a2*lat+b2*lon+c2
        * First  point is (1,1)   i.e. Near Range, First Lane <==> Lat[0,0],Lon[0,0]
        * Second point is (nx,1)  i.e. Far Range, First Lane <==> Lat[0,-1],Lon[0,-1]
        * Third  point is (1,ny)  i.e. Near Range, Last Lane <==> Lat[-1,0],Lon[-1,0]
        * Fourth point is (nx,ny) i.e. Far Range, Last Lane <==> Lat[-1,-1],Lon[-1,-1] 
    '''

    # Size
    ny, nx = lon.shape

    # Mapping function
    A=np.array([ [lat[0,0], lon[0,0], 1.],
                 [lat[0,-1], lon[0,-1], 1.],
                 [lat[-1,0], lon[-1,0], 1.],
                 [lat[-1,-1], lon[-1,-1], 1.]])
    b=np.array([[1., 1.],[nx, 1.],[1., ny],[nx,ny]])
    mfcn=np.linalg.lstsq(A,b)[0]

    #Get grid points xi yi coordinates from this mapping function
    A=np.vstack([latlist, lonlist, np.ones(len(lonlist),)]).T
    xi=np.dot(A,mfcn[:,0])
    yi=np.dot(A,mfcn[:,1])
    
    # Plot y/n
    if plotflag:
        from matplotlib import pyplot as plt, patches

        plt.figure()
        plt.subplot(211)
        plt.scatter(lonlist,latlist,s=8,c=np.cumsum(np.ones(len(lonlist),)))
        xline=[lon[0,0],lon[0,-1],lon[-1,-1],lon[-1,0],lon[0,0]]
        yline=[lat[0,0],lat[0,-1],lat[-1,-1],lat[-1,0],lat[0,0]]
        plt.plot(xline,yline,'-r')
        plt.title('Area of interest and {} stations used'.format(len(lonlist)))
        plt.xlabel('Longitude')
        plt.ylabel('Latitude')

        plt.subplot(212)
        plt.scatter(xi,yi,s=8,c=np.cumsum(np.ones(len(lonlist),)))
        p = patches.Rectangle((1,1),lon.shape[1],lon.shape[0],
                                 edgecolor="Red",fill=False)
        plt.gca().add_patch(p)
        plt.title('Area of interest in Radar Geometry')
        plt.xlabel('Range')
        plt.ylabel('Azimuth')
        plt.show()

    # All done
    return np.array(xi).astype(float), np.array(yi).astype(float)


def glob2rdr(nx,ny,lat,lon,latl,lonl,plotflag='n'):
    '''Transfert these latlon coordinates into radar geometry (xi,yi) with a 
       simple linear transformation given the first pixel and the pixel 
       spacing of the simulation.

    Args:
        * nx   (int)        : Number of range bins.
        * ny   (int)        : Number of azimuth lines.
        * lat  (np.ndarray) : Array of latitudes of the corners
        * lon  (np.ndarray) : Array of longitudes of the corners
        * latl (np.ndarray) : Latitudes of the stations.
        * lonl (np.ndarray) : Longitudes of the stations.

    Kwargs:
        * plotflag (bool)   : Plot the stations distribution.

    Returns:
        * xi (np.ndarray)   : Position of stations in range.
        * yi (np.ndarray)   : Position of stations in azimuth.

    .. note::
        Mapping function is :
        * range = a1*lat+b1*lon+c1
        * azimu = a2*lat+b2*lon+c2
        * First  point is (1,1)   i.e. Near Range, First Lane <==> Lat[1],Lon[1]
        * Second point is (nx,1)  i.e. Far Range, First Lane <==> Lat[2],Lon[2]
        * Third  point is (1,ny)  i.e. Near Range, Last Lane <==> Lat[3],Lon[3]
        * Fourth point is (nx,ny) i.e. Far Range, Last Lane <==> Lat[4],Lon[4] '''

    # Mapping function
    A = np.hstack([lat,lon,np.ones((4,1))])
    b = np.array([[1, 1],[nx, 1],[1, ny],[nx,ny]])
    mfcn = np.linalg.lstsq(A,b,rcond=None)[0]

    #Get grid points xi yi coordinates from this mapping function
    nstn = latl.size
    A = np.array([latl.reshape(-1,1), lonl.reshape(-1,1), np.ones((nstn,1))]).T
    xi = np.dot(A,mfcn[:,0]).reshape(latl.shape)
    yi = np.dot(A,mfcn[:,1]).reshape(latl.shape)

    if plotflag in ('y','Y'):
        from matplotlib import pyplot as plt, patches

        plt.figure(1)
        plt.subplot(211)
        plt.scatter(lonl,latl,s=8,c='k');
        xline=[lon[0],lon[1],lon[3],lon[2],lon[0]]
        yline=[lat[0],lat[1],lat[3],lat[2],lat[0]]
        plt.plot(xline,yline,'-r')
        plt.title('Area of interest and %d stations used'%(nstn))
        plt.xlabel('Longitude')
        plt.ylabel('Latitude')

        plt.subplot(212)
        plt.scatter(xi,yi,s=8,c='k')
        p = patches.Rectangle((1,1),nx,ny,edgecolor="Red",fill=False)
        plt.gca().add_patch(p)
        plt.title('Area of interest in Radar Geometry')
        plt.xlabel('Range')
        plt.ylabel('Azimuth')
        plt.show()

    return xi,yi

############### obsolete: simple geo2radar - end ######



############### obsolete: simple radar2geo ############

def rdr2glob(wid,lgt,lat,lon,x,y,plotflag='n'):   #nx,ny,lat,lon,latl,lonl,plotflag='n'):
    '''Transfert these radar geometry (x,y) coordinates in lat/lon coordinates with a 
    simple linear transformation given the image width/length and the lat/lon coordinates
    of the 4 corners

    Args:
            * wid  (int)        : Width of the image (i.e. number of range bins)
            * lgt  (int)        : Length of the image (i.e. number of azimuth lines)
            * lat  (np.ndarray) : Array of latitudes of the corners
            * lon  (np.ndarray) : Array of longitudes of the corners
            * rang (np.ndarray) : Range of the points to transfert
            * azim (np.ndarray) : Azimuth of the points to transfert

    Kwargs:
            * plotflag (bool)   : Plot the stations distribution.

    Returns:
            * loni (np.ndarray) : Longitude of the points.
            * lati (np.ndarray) : Latitude of the points.

    .. note::
            Mapping function is :
            * lat = a1*rang + b1*azim + c1
            * lon = a2*rang + b2*azim + c2
            * First  point is (1,1)   i.e. Near Range, First Lane <==> Lat[1],Lon[1]
            * Second point is (nx,1)  i.e. Far Range, First Lane <==> Lat[2],Lon[2]
            * Third  point is (1,ny)  i.e. Near Range, Last Lane <==> Lat[3],Lon[3]
            * Fourth point is (nx,ny) i.e. Far Range, Last Lane <==> Lat[4],Lon[4] '''

    A = np.array([[1, 1, 1.],[wid, 1, 1.],[1, lgt, 1.],[wid, lgt, 1.]])
    b = np.hstack((lat,lon))
    mfcn = np.linalg.lstsq(A,b)[0]

    #Get grid points xi yi coordinates from this mapping function
    nstn = len(x)
    A = np.array([x, y, np.ones((nstn,1))]).T
    lati = np.dot(A,mfcn[:,0])
    loni = np.dot(A,mfcn[:,1])

    if plotflag in ('y','Y'):
        from matplotlib import pyplot as plt, patches

        plt.figure(1)
        plt.subplot(211)
        plt.scatter(lon,lat,s=8,c='k');
        xline=[lon[0],lon[1],lon[3],lon[2],lon[0]]
        yline=[lat[0],lat[1],lat[3],lat[2],lat[0]]
        plt.scatter(loni,lati,s=8,c='r')
        plt.plot(xline,yline,'-r')
        plt.title('Area of interest in geographic coordinates')
        plt.xlabel('Longitude')
        plt.ylabel('Latitude')

        plt.subplot(212)
        plt.scatter(x,y,s=8,c='k')
        p = patches.Rectangle((1,1),wid,lgt,edgecolor="Red",fill=False)
        plt.gca().add_patch(p)
        plt.title('Area of interest in Radar Geometry')
        plt.xlabel('Range')
        plt.ylabel('Azimuth')
        plt.show()

    return loni,lati

############### obsolete: simple radar2geo - end ######



############### Simple progress bar ###################

class ProgressBar:
    """ Creates a text-based progress bar. Call the object with 
    the simple `print'command to see the progress bar, which looks 
    something like this:
    [=======> 22% ]
    You may specify the progress bar's width, min and max values on init.
    
    .. note::
        Code originally from http://code.activestate.com/recipes/168639/"""

    def __init__(self, minValue = 0, maxValue = 100, totalWidth=80):
        self.progBar = "[]" # This holds the progress bar string
        self.min = minValue
        self.max = maxValue
        self.span = maxValue - minValue
        self.width = totalWidth
        self.reset()

    def reset(self):
        self.start_time = time.time()
        self.amount = 0 # When amount == max, we are 100% done
        self.updateAmount(0) # Build progress bar string

    def updateAmount(self, newAmount = 0):
        """ Update the progress bar with the new amount (with min and max
        values set at initialization; if it is over or under, it takes the
        min or max value as a default. """
        if newAmount < self.min:
            newAmount = self.min
        if newAmount > self.max:
            newAmount = self.max
        self.amount = newAmount

        # Figure out the new percent done, round to an integer
        diffFromMin = float(self.amount - self.min)
        percentDone = (diffFromMin / float(self.span)) * 100.0
        percentDone = int(np.round(percentDone))

        # Figure out how many hash bars the percentage should be
        allFull = self.width - 2 - 18
        numHashes = (percentDone / 100.0) * allFull
        numHashes = int(np.round(numHashes))

        # Build a progress bar with an arrow of equal signs; special cases for
        # empty and full
        if numHashes == 0:
            self.progBar = '[>%s]' % (' '*(allFull-1))
        elif numHashes == allFull:
            self.progBar = '[%s]' % ('='*allFull)
        else:
            self.progBar = '[%s>%s]' % ('='*(numHashes-1),
                            ' '*(allFull-numHashes))
            # figure out where to put the percentage, roughly centered
            percentPlace = (len(self.progBar) // 2) - len(str(percentDone))
            percentString = ' ' + str(percentDone) + '% '
            elapsed_time = time.time() - self.start_time
            # slice the percentage into the bar
            self.progBar = ''.join([self.progBar[0:percentPlace],
                                    percentString,
                                    self.progBar[percentPlace+len(percentString):]])
            if percentDone > 0:
                self.progBar += ' %6ds / %6ds' % (int(elapsed_time),int(elapsed_time*(100.//percentDone-1)))

    def update(self, value, every=1):
        """ Updates the amount, and writes to stdout. Prints a
         carriage return first, so it will overwrite the current
          line in stdout."""
        if value % every == 0 or value >= self.max:
            self.updateAmount(newAmount=value)
            sys.stdout.write('\r' + self.progBar)
            sys.stdout.flush()

    def close(self):
        """Prints a blank space at the end to ensure proper printing
        of future statements."""
        print(' ')

############### Simple progress bar - end #############



############################################################
# Program is part of PyAPS                                 #
# Copyright 2012, by the California Institute of Technology#
# Contact: earthdef@gps.caltech.edu                        #
# Modified by A. Benoit and R. Jolivet 2019                #
# Ecole Normale Superieure, Paris                          #
# Contact: insar@geologie.ens.fr                           #
############################################################
import numpy as np

#############Clausius-Clapeyron for ECMWF as used in Jolivet et al 2011#
def cc_eraorig(tmp,cdic):
    '''This routine takes temperature profiles and returns Saturation water vapor
    partial pressure using the Clausius-Clapeyron law applied in Jolivet et al. 2011,
    GRL, doi:10.1029/2011GL048757. It can be used in case you are using Relative.
    Humidity from ECMWF models.

    Args:
        * tmp  (np.ndarray) : Temperature
        * cdic (dict)       : Dictionnary of constants

    Returns:
        * esat (np.ndarray) : Saturation water vapor partial pressure.
    '''

    a1w=cdic['a1w']
    T3=cdic['T3']
    Rv=cdic['Rv']

    esat=a1w*np.exp( (2.5e6/Rv) * ( (1/T3) - (1/tmp) ) )

    return esat
###############Completed CC_ERAORIG#####################################


##############Read Input text file #####################################
def read_eratxt(fname,cdic):
    '''Read ECMWF files from 0.75 degree grid similar to Romain Jolivet's Delay Package.

    Args:
        * fname    (str)       :  Path to the delay file
        * cdic     (float)     :  Dictionary of constants

    Kwargs:
        * humidity (str)       : Specific ('Q') or relative humidity ('R').

    Returns:
        * lvls     (np.ndarray): Pressure levels
        * latlist  (np.ndarray): Latitudes of the stations
        * lonlist  (np.ndarray): Longitudes of the stations
        * gph      (np.ndarray): Geopotential height
        * tmp      (np.ndarray): Temperature
        * vpr      (np.ndarray): Vapor pressure

    .. note::
        Uses cc_eraorig by default.
    '''

    lvls=[]
    latlist=[]
    lonlist=[]
    gpht=[]
    tmpt=[]
    reht=[]

    g=cdic['g']

    f=open(fname,'r')
    tmp=f.readlines()
    i=0
    nstn=0
    maxloop=int(len(tmp))
    while i<maxloop:
        if (tmp[i][0]=='-'):
            nstn=nstn+1
            lonlat=tmp[i+3].rsplit(' ')
            lonlist.append(float(lonlat[3]))
            latlist.append(float(lonlat[9]))
            i=i+5
            new='y'
        else:
            if new in ('y'):
                n=1
                val=tmp[i].rsplit(' ')
                gpht.append(float(val[0]))
                lvls.append(float(val[1]))
                tmpt.append(float(val[2]))
                reht.append(float(val[3]))
                i=i+1
                new='n'
            else:
                n=n+1
                val=tmp[i].rsplit(' ')
                gpht.append(float(val[0]))
                lvls.append(float(val[1]))
                tmpt.append(float(val[2]))
                reht.append(float(val[3]))
                i=i+1

    gpht=np.array(gpht)/g
    gph=np.flipud(gpht.reshape((n,nstn),order='F'))
    del gpht

    tmpt=np.array(tmpt)
    esat=cc_eraorig(tmpt,cdic)
    tmp=np.flipud(tmpt.reshape((n,nstn),order='F'))
    del tmpt

    vprt=(np.array(reht)/100.)*esat
    vpr=np.flipud(vprt.reshape((n,nstn),order='F'))
    del vprt
    del esat

    lvls=np.flipud(np.array(lvls))
    lvls=lvls[0:n]

    lonlist=np.array(lonlist)
    latlist=np.array(latlist)

    return lvls,latlist,lonlist,gph,tmp,vpr



from pyhdf.HDF import *
from pyhdf.SD import *
from pyhdf.V import *
from pyhdf.VS import *
import numpy as np
import netCDF4


def get_merra(fname,minlat,maxlat,minlon,maxlon,cdic,verbose=False):
    '''Read data from MERRA hdf file. Note that the Lon values
       should be between [0-360]. Hdf file with weather model
       data can be downloaded from
       http://disc.sci.gsfc.nasa.gov/daac-bin/FTPSubset.pl

    Args:
        * fname   (str)       : Path to the grib file
        * minlat  (float)     : Minimum latitude
        * maxlat  (float)     : Maximum latitude
        * minlon  (float)     : Minimum longitude
        * maxlon  (float)     : Maximum longitude
        * cdic    (float)     : Dictionary of constants

    Returns:
        * lvls    (np.ndarray): Pressure levels
        * latlist (np.ndarray): Latitudes of the stations
        * lonlist (np.ndarray): Longitudes of the stations
        * gph     (np.ndarray): Geopotential height
        * tmp     (np.ndarray): Temperature
        * vpr     (np.ndarray): Vapor pressure

        '''

    if fname[-3:]=='hdf':
        print(minlat)
        print(maxlat)
        print(minlon)
        print(maxlon)

        # Read the hdf file
        file = SD(fname)
        if verbose:
            print('PROGRESS: READING HDF FILE')
        lvl = file.select('levels')
        # Pressure levels are from lowest to highest
        rlvls = lvl.get()
        lvls = []
        # Reverse the pressure levels to be consistent with other GAMss
        for i in range(len(rlvls)):
            index = len(rlvls) - i - 1
            lvls.append(rlvls[index])
        nlvls = len(lvls)
        lvls = np.array(lvls)

        alpha = cdic['Rv']/cdic['Rd']

        # Select latitutde and longitude
        lat = file.select('latitude')
        lon = file.select('longitude')
        lats = lat.get()
        lons = lon.get()
        mask1 = (lats > minlat) & (lats < maxlat)
        mask2 = (lons > minlon) & (lons < maxlon)
        [ii] = np.where(mask1 == True)
        [jj] = np.where(mask2 == True)
        del mask1
        del mask2
        iimemo = []
        for m in range(len(ii)):
            for i in range(len(jj)):
                iimemo.append(ii[m])
        jjmemo = []
        for i in range(len(ii)):
            jjmemo.append(jj)
        jjmemo = np.array(jjmemo)
        jjmemo = jjmemo.flatten()
        iimemo = np.array(iimemo)
        ii = iimemo
        jj = jjmemo
        latlist = lats[ii]
        lonlist = lons[jj]
        nstn = len(latlist)

        # Create arrays for 3D storage
        gph = np.zeros((nlvls, nstn))     #Potential height
        tmp = gph.copy()                  #Temperature
        vpr = gph.copy()                  #Vapor pressure
        if verbose:
            print('Number of stations:', nstn)

        # Get data from files
        h = file.select('h')
        height = h.get()[0]
        qv = file.select('qv')
        humidity = qv.get()[0]
        sp = file.select('ps')
        spressure = sp.get()[0]
        t = file.select('t')
        temp = t.get()[0]

        # Lvls are in hecto pascals, convert to pascals
        lvls = 100.0*lvls

        # Reverse altitude
        for i in range(nlvls):
            index = nlvls - i - 1
            gph[i,:] = height[index][ii,jj]

        idx = np.zeros(nstn)
        for i in range(nstn):
            for m in range(nlvls):
                if spressure[ii[i]][jj[i]] > lvls[m]:
                    idx[i] = m

        # extrapolation of temperature and humidity data at pressure levels under surface at each grid point
        tk = np.zeros(nstn)
        tb = np.zeros(nstn)
        for i in range(nstn):
            t = temp[:,ii[i],jj[i]]
            x = [lvls[idx[i]],lvls[idx[i] -1 ]]
            y = [t[nlvls - idx[i] - 1],t[nlvls - idx[i] ]]
            coef = np.polyfit(x,y,1)
            tk[i] = coef[0]
            tb[i] = coef[1]

        hk = np.zeros(nstn)
        hb = np.zeros(nstn)
        for i in range(nstn):
            hum = humidity[:,ii[i],jj[i]]
            x = [lvls[idx[i]],lvls[idx[i] -1 ]]
            y = [hum[nlvls - idx[i] - 1],hum[nlvls - idx[i]]]
            coef = np.polyfit(x,y,1)
            hk[i] = coef[0]
            hb[i] = coef[1]


        #fill out the tmp and vpr array
        for i in range(nstn):
            ind = int(idx[i]+1)
            for n in range(ind):
                tmp[n,i] = temp[nlvls - 1 - n,ii[i],jj[i]]
            exl = nlvls - ind
            for m in range(exl):
                tmp[ind+m,i] =  tk[i]*lvls[ind+m] + tb[i]

        for i in range(nstn):
            ind = int(idx[i]+1)
            for n in range(ind):
                vpr[n,i] = humidity[nlvls - 1- n,ii[i],jj[i]]
            exl = nlvls - ind
            for m in range(exl):
                vpr[ind+m,i] =  hk[i]*lvls[ind+m] + hb[i]

        memo = list(vpr)
        memo = np.array(memo)
        for i in range(nlvls):
            vpr[i,:] = memo[i,:]*lvls[i]*alpha/(1+(alpha - 1)*memo[i,:])

        # Close the hdf file
        file.end()

    if fname[-3:]=='nc4':
        # Read the netcdf file
        file = netCDF4.Dataset(fname)
        if verbose:
            print('PROGRESS: READING netcdf FILE')
        ncv = file.variables
        # Pressure levels are from lowest to highest
        rlvls = ncv['lev'][:]
        lvls = []
        # Reverse the pressure levels to be consistent with other GAMs
        for i in range(len(rlvls)):
            index = len(rlvls) - i - 1
            lvls.append(rlvls[index])
        nlvls = len(lvls)
        lvls = np.array(lvls)

        alpha = cdic['Rv']/cdic['Rd']

        # Select latitutde and longitude
        lats = ncv['lat'][:]
        lons = ncv['lon'][:]
        mask1 = (lats > minlat) & (lats < maxlat)
        mask2 = (lons > minlon) & (lons < maxlon)
        [ii] = np.where(mask1 == True)
        [jj] = np.where(mask2 == True)
        del mask1
        del mask2
        iimemo = []
        for m in range(len(ii)):
            for i in range(len(jj)):
                iimemo.append(ii[m])
        jjmemo = []
        for i in range(len(ii)):
            jjmemo.append(jj)
        jjmemo = np.array(jjmemo)
        jjmemo = jjmemo.flatten()
        iimemo = np.array(iimemo)
        ii = iimemo
        jj = jjmemo
        latlist = lats[ii]
        lonlist = lons[jj]
        nstn = len(latlist)

        # Create arrays for 3D storage
        gph = np.zeros((nlvls, nstn))     #Potential height
        tmp = gph.copy()                  #Temperature
        vpr = gph.copy()                  #Vapor pressure
        if verbose:
            print('Number of stations:', nstn)

        # Get data from files
        height = ncv['H'][:]
        height = height[0]
        humidity = ncv['QV'][:]
        humidity = humidity[0]
        spressure = ncv['PS'][:]
        spressure = spressure[0]
        temp = ncv['T'][:]
        temp = temp[0]

        # Lvls are in hecto pascals, convert to pascals
        lvls = 100.0*lvls

        # Reverse altitude
        for i in range(nlvls):
            index = nlvls - i - 1
            gph[i,:] = height[index][ii,jj]

        idx = np.zeros(nstn)
        for i in range(nstn):
            for m in range(nlvls):
                if spressure[ii[i]][jj[i]] > lvls[m]:
                    idx[i] = m

        # extrapolation of temperature and humidity data at pressure levels under surface at each grid point
        tk = np.zeros(nstn)
        tb = np.zeros(nstn)
        for i in range(nstn):
            t = temp[:,ii[i],jj[i]]
            x = [lvls[idx[i]],lvls[idx[i] -1 ]]
            y = [t[nlvls - idx[i] - 1],t[nlvls - idx[i] ]]
            coef = np.polyfit(x,y,1)
            tk[i] = coef[0]
            tb[i] = coef[1]

        hk = np.zeros(nstn)
        hb = np.zeros(nstn)
        for i in range(nstn):
            hum = humidity[:,ii[i],jj[i]]
            x = [lvls[idx[i]],lvls[idx[i] -1 ]]
            y = [hum[nlvls - idx[i] - 1],hum[nlvls - idx[i]]]
            coef = np.polyfit(x,y,1)
            hk[i] = coef[0]
            hb[i] = coef[1]


        #fill out the tmp and vpr array
        for i in range(nstn):
            ind = int(idx[i]+1)
            for n in range(ind):
                tmp[n,i] = temp[nlvls - 1 - n,ii[i],jj[i]]
            exl = nlvls - ind
            for m in range(exl):
                tmp[ind+m,i] =  tk[i]*lvls[ind+m] + tb[i]

        for i in range(nstn):
            ind = int(idx[i]+1)
            for n in range(ind):
                vpr[n,i] = humidity[nlvls - 1- n,ii[i],jj[i]]
            exl = nlvls - ind
            for m in range(exl):
                vpr[ind+m,i] =  hk[i]*lvls[ind+m] + hb[i]
        memo = list(vpr)
        memo = np.array(memo)
        for i in range(nlvls):
            vpr[i,:] = memo[i,:]*lvls[i]*alpha/(1+(alpha - 1)*memo[i,:])

        # Close the netcdf file
        file.close()

    # Send data
    return lvls,latlist,lonlist,gph,tmp,vpr


############################################################
# Program is part of PyAPS                                 #
# Copyright 2012, by the California Institute of Technology#
# Contact: earthdef@gps.caltech.edu                        #
############################################################



#!/usr/bin/env python
#
# (C) Copyright 2012-2013 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
# In applying this licence, ECMWF does not waive the privileges and immunities 
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

from ecmwfapi import ECMWFDataServer

# To run this example, you need an API key 
# available from https://api.ecmwf.int/v1/key/

server = ECMWFDataServer()
server.retrieve({
    'dataset' : "tigge",
    'step'    : "24",
    'number'  : "all",
    'levtype' : "sl",
    'date'    : "20071001",
    'time'    : "00",
    'origin'  : "all",
    'type'    : "pf",
    'param'   : "tp",
    'area'    : "70/-130/30/-60",
    'grid'    : "2/2",
    'target'  : "data.grib"
})



############################################################
# Program is part of PyAPS                                 #
# Copyright 2012, by the California Institute of Technology#
# Contact: earthdef@gps.caltech.edu                        #
# Modified by A. Benoit and R. Jolivet 2019                #
# Ecole Normale Superieure, Paris                          #
# Contact: insar@geologie.ens.fr                           #
############################################################
'''Script to automatically download weather model data corresponding to our interferograms.'''

import PyAPS
import tsinsar as ts
import datetime as dt

dates,sat,bperp = ts.load_list('ifg.list')
days,usat,Jmat = ts.ConnectMatrix(dates,sat)


daylist = []
for k in days:
    dobj = dt.date.fromordinal(int(k))
    strobj = '%4d%02d%02d'%(dobj.year,dobj.month,dobj.day)
    daylist.append(strobj)

PyAPS.ECMWFdload(daylist,'18','./Atmos/ECMWF/')

#PyAPS.NARRdload(daylist,'18','./Atmos/NARR/')
#PyAPS.ERAdload(daylist,'18','./Atmos/ERA')



#!/usr/bin/env python
import pyaps3 as pa
import numpy as np
import matplotlib.pyplot as plt
import sys

print('------------------------------------------------')
print('You are using PyAPS from %s'%pa.__file__)
print('------------------------------------------------')

print('Testing Download Methods')
print('Testing ECMWF Downloads')
pa.ECMWFdload(['20150503'],'00','./ECMWF','interim','fc')

print 'Testing NARR Downloads'
pa.NARRdload(['20040526','20030426'],'12','./NARR/')

print 'Testing ERA@UCAR Downloads'
pa.ERAdload(['20050708','20041201'],'12','./ERA/')

print('Downloads OK')
print('------------------------------------------------')
print('------------------------------------------------')

print('Testing ECMWF in Radar geometry, with a RMG dem')
aps1 = pa.PyAPS_rdr('ECMWF/ERA-Int_20150503_12.grb','/data/angel/TestISCE/stack/INTERFEROGRAMS/hgt.9alks_27rlks.full.rdr',grib='ECMWF',verb=True)
aps2 = pa.PyAPS_rdr('ECMWF/ERA-Int_20040526_12.grb','dem_16rlks.hgt',grib='ECMWF',verb=True)

print('With Lat Lon files')
phs1 = np.zeros((aps1.ny,aps1.nx))
phs2 = np.zeros((aps2.ny,aps2.nx))

aps1.getgeodelay(phs1,inc='/data/angel/TestISCE/stack/INTERFEROGRAMS/losFull.9alks_27rlks.rdr',wvl=0.05546576,lat='/data/angel/TestISCE/stack/INTERFEROGRAMS/latFull.9alks_27rlks.rdr',lon='/data/angel/TestISCE/stack/INTERFEROGRAMS/lonFull.9alks_27rlks.rdr')
aps2.getgeodelay(phs2,inc=23.0,wvl=0.056,lat='lat.flt',lon='lon.flt')

LLphs = phs2-phs1
plt.subplot(121)
plt.imshow(LLphs[10:600,10:350])
plt.title('With latlon file')
plt.colorbar()
plt.show()

print('Testing ECMWF in Radar geometry, with a FLT dem')
aps1 = pa.PyAPS_rdr('ECMWF/ERA-Int_20030426_12.grb','DEM.flt',grib='ECMWF',demfmt='HGT')
aps2 = pa.PyAPS_rdr('ECMWF/ERA-Int_20040526_12.grb','DEM.flt',grib='ECMWF',demfmt='HGT')

print('With Lat Lon files')
phs1 = np.zeros((aps1.ny,aps1.nx))
phs2 = np.zeros((aps2.ny,aps2.nx))

aps1.getgeodelay(phs1,inc=23.0,wvl=0.056,lat='lat.flt',lon='lon.flt')
aps2.getgeodelay(phs2,inc=23.0,wvl=0.056,lat='lat.flt',lon='lon.flt')

LLphs = phs2-phs1
plt.subplot(121)
plt.imshow(LLphs[10:600,10:350])
plt.title('With latlon file')
plt.colorbar()
plt.show()

print('Testing ECMWF in Geographic coordinates, with a FLT dem')
aps1 = pa.PyAPS_geo('ECMWF/ERA-Int_20030426_12.grb','Dem_crop.dem',grib='ECMWF')
aps2 = pa.PyAPS_geo('ECMWF/ERA-Int_20040526_12.grb','Dem_crop.dem',grib='ECMWF')

print('With Lat Lon files')
phs1 = np.zeros((aps1.ny,aps1.nx))
phs2 = np.zeros((aps2.ny,aps2.nx))

aps1.getdelay(phs1,inc=23.0,wvl=0.056)
aps2.getdelay(phs2,inc=23.0,wvl=0.056)

del phs1
del phs2

print('ECMWF OK')
print('------------------------------------------------')
print('------------------------------------------------')

sys.exit(1)

print('Testing NARR in Radar geometry, with a RMG dem')
aps1 = pa.PyAPS_rdr('NARR/narr-a_221_20030426_1200_000.grb','hgt.flt',grib='NARR',demfmt='HGT')
aps2 = pa.PyAPS_rdr('NARR/narr-a_221_20040526_1200_000.grb','hgt.flt',grib='NARR',demfmt='HGT')
print('With Lat Lon files')
phs1 = np.zeros((aps1.ny,aps1.nx))
phs2 = np.zeros((aps2.ny,aps2.nx))

aps1.getgeodelay(phs1,inc=23.0,wvl=0.056,lat='lat2.flt',lon='lon2.flt')
aps2.getgeodelay(phs2,inc=23.0,wvl=0.056,lat='lat2.flt',lon='lon2.flt')

phs = phs2-phs1
plt.imshow(phs)
plt.colorbar()
plt.show()

print('------------------------------------------------')
print('------------------------------------------------')

print('All Test OK')



############################################################
# Program is part of PyAPS                                 #
# Copyright 2012, by the California Institute of Technology#
# Contact: earthdef@gps.caltech.edu                        #
############################################################



#!/usr/bin/env python
#
# (C) Copyright 2012-2013 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
# In applying this licence, ECMWF does not waive the privileges and immunities 
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

from setuptools import setup, find_packages

import ecmwfapi


setup(
    name="ecmwf-api-client",
    version=ecmwfapi.__version__,
    description=ecmwfapi.__doc__,
    author="ECMWF",
    author_email="software.support@ecmwf.int",
    url="https://software.ecmwf.int/stash/projects/PRDEL/repos/ecmwf-api-client/browse",

    # entry_points={
    #     "console_scripts": [
    #         "mars = XXX:main",
    #     ],
    # },

    packages=find_packages(),
    zip_safe=False,
)



############################################################
# Program is part of PyAPS                                 #
# Copyright 2012, by the California Institute of Technology#
# Contact: earthdef@gps.caltech.edu                        #
# Modified by A. Benoit and R. Jolivet 2019                #
# Ecole Normale Superieure, Paris                          #
# Contact: insar@geologie.ens.fr                           #
############################################################
import numpy as np
import pygrib


#############Clausis-Clayperon for ECMWF###########################
def cc_era(tmp,cdic):
    '''Clausius Clayperon law used by ERA Interim.

    Args:
        * tmp  (np.ndarray) : Temperature.
        * cdic (dict)       : Dictionary of constants

    Returns:
        * esat (np.ndarray) : Water vapor saturation partial pressure.'''


    a1w = cdic['a1w']
    a3w = cdic['a3w']
    a4w = cdic['a4w']
    a1i = cdic['a1i']
    a3i = cdic['a3i']
    a4i = cdic['a4i']
    T3  = cdic['T3']
    Ti  = cdic['Ti'] 

    esatw = a1w*np.exp(a3w*(tmp-T3)/(tmp-a4w))
    esati = a1i*np.exp(a3i*(tmp-T3)/(tmp-a4i))
    esat = esati.copy()
    for k in range(len(tmp)):
        if (tmp[k] >= T3):
            esat[k] = esatw[k]
        elif (tmp[k] <= Ti):
            esat[k] = esati[k]
        else:
            wgt = (tmp[k]-Ti)/(T3-Ti)
            esat[k] = esati[k] + (esatw[k]-esati[k])*wgt*wgt

    return esat


########Read in ERA data from a given ERA Interim file##################
def get_era(fname,minlat,maxlat,minlon,maxlon,cdic, humidity='Q',verbose=False):
    '''Read data from ERA interim grib file. 
    Note that the Lon values should be between [0-360]. 
    GRB file with weather model data can be downloaded from http://rda.ucar.edu/datasets/ds627.0/

    Args:
        * fname   (str)       : Path to the grib file
        * minlat  (float)     : Minimum latitude
        * maxlat  (float)     : Maximum latitude
        * minlon  (float)     : Minimum longitude
        * maxlon  (float)     : Maximum longitude
        * cdic    (float)     : Dictionary of constants
    
    Kwargs:
        * humidity(str)       : Specific ('Q') or relative humidity ('R').

    Returns:
        * lvls    (np.ndarray): Pressure levels
        * latlist (np.ndarray): Latitudes of the stations
        * lonlist (np.ndarray): Longitudes of the stations
        * gph     (np.ndarray): Geopotential height
        * tmp     (np.ndarray): Temperature
        * vpr     (np.ndarray): Vapor pressure

    .. note::
        Uses cc_era by default.
        '''
    
    assert humidity in ('Q','R'), 'Undefined humidity field in get_era.'
    if verbose:
        print('PROGRESS: READING GRIB FILE {}'.format(fname))
    lvls = np.array([1, 2, 3, 5, 7, 10, 20, 30, 50, 70, 100, 125, 150, 175, 
                     200, 225, 250, 300, 350, 400, 450, 500, 550, 600, 650, 700, 750, 775, 
                     800, 825, 850, 875, 900, 925, 950, 975, 1000])
    nlvls = len(lvls)

    alpha = cdic['Rv']/cdic['Rd']
    gphind = np.arange(nlvls)*12+1
    
    grbs = pygrib.open(fname)
    grbs.seek(gphind[0])
    grb=grbs.read(1)[0]
    lats, lons = grb.latlons()
    g = cdic['g']
    mask = (lats > minlat) & (lats < maxlat) & (lons > minlon) & (lons < maxlon)
    [ii,jj] = np.where(mask == True)
    del mask
    latlist = lats[ii,jj]
    lonlist = lons[ii,jj]
    nstn = len(ii)
    
    ####Create arrays for 3D storage
    gph = np.zeros((nlvls, nstn))     #Potential height
    tmp = gph.copy()                  #Temperature
    vpr = gph.copy()                  #Vapor pressure
    if verbose:
        print('Number of stations:', nstn)

    lvls = 100.0*lvls                 #Conversion to absolute pressure
    for i in range(nlvls):
        grbs.seek(gphind[i])          #Reading potential height.
        grb = grbs.read(3)
        val = grb[0].values
        gph[i,:] = val[ii,jj]/g

        val = grb[1].values           #Reading temperature
        temp = val[ii,jj]
        tmp[i,:] = temp

        if humidity in ('R'):
            esat = cc_era(temp,cdic)       
            grbs.seek(gphind[i]+6)
            grb = grbs.read(1)
            val = grb[0].values
            temp = val[ii,jj]/100.0
            vpr[i,:] = temp*esat
                
        elif humidity in ('Q'):
            val = grb[2].values       #Specific humidity
            temp = val[ii,jj]
            vpr[i,:] = temp*lvls[i]*alpha/(1+(alpha - 1)*temp)
        
        else:
             assert 1==0, 'Undefined Humidity in get_era().' 

    return lvls,latlist,lonlist,gph,tmp,vpr
###############Completed GET_ERA########################################


########Read in ERA data from a given ERA Interim file##################
def get_ecmwf(model,fname,minlat,maxlat,minlon,maxlon,cdic, humidity='Q',verbose=False):
    '''Read data from ERA Interim, ERA-5 or HRES grib file. Note that Lon values should be between [-180, 180].
    Modified by A. Benoit, January 2019.

    Args:
        * model    (str)       : Model used (ERA5, ERAINT or HRES)
        * fname    (str)       : Path to the grib file
        * minlat   (float)     : Minimum latitude
        * maxlat   (float)     : Maximum latitude
        * minlon   (float)     : Minimum longitude
        * maxlon   (float)     : Maximum longitude
        * cdic     (float)     : Dictionary of constants
    
    Kwargs:
        * humidity (str)       : Specific ('Q') or relative humidity ('R').

    Returns:
        * lvls     (np.ndarray): Pressure levels
        * latlist  (np.ndarray): Latitudes of the stations
        * lonlist  (np.ndarray): Longitudes of the stations
        * gph      (np.ndarray): Geopotential height
        * tmp      (np.ndarray): Temperature
        * vpr      (np.ndarray): Vapor pressure

    .. note::
        Uses cc_era by default.
    '''

    assert humidity in ('Q','R'), 'Undefined humidity field in get_era.'
    assert model in ('ERA5', 'ERAINT','HRES'), 'Model not recognized.'
    if verbose:
        print('PROGRESS: READING GRIB FILE')
    if model in 'HRES':
        if verbose:
            print('INFO: USING PRESSURE LEVELS OF HRES DATA')
        lvls = np.array([1, 2, 3, 5, 7, 10, 20, 30, 50, 70, 100, 150, 
                         200, 250, 300, 400, 500, 600, 700,
                         800, 850, 900, 925, 950, 1000])
    else:
        if verbose:
            print('INFO: USING PRESSURE LEVELS OF ERA-INT OR ERA-5 DATA')
        lvls = np.array([1, 2, 3, 5, 7, 10, 20, 30, 50, 70, 100, 125, 150, 175, 
                         200, 225, 250, 300, 350, 400, 450, 500, 550, 600, 650, 700, 750, 775,
                         800, 825, 850, 875, 900, 925, 950, 975, 1000])
    nlvls = len(lvls)

    alpha = cdic['Rv']/cdic['Rd']
    gphind = np.arange(nlvls)*3

    grbs = pygrib.open(fname)
    grbs.seek(gphind[0])
    grb = grbs.read(1)[0]
    lats, lons = grb.latlons()
    #if model == 'ERA5':
    #    lons[lons < 0.] += 360.
    g = cdic['g']

    #extract indices 
    mask = ((lats > minlat) & (lats < maxlat)) & ((lons > minlon) & (lons < maxlon))
    uu = [i for i in list(range(np.shape(mask)[0])) if any(mask[i,:])]
    vv = [j for j in list(range(np.shape(mask)[1])) if any(mask[:,j])]
    
    latlist = lats[uu,:][:,vv]
    lonlist = lons[uu,:][:,vv]
    nlat, nlon = latlist.shape

    ####Create arrays for 3D storage
    gph = np.zeros((nlvls, nlat, nlon))   #Potential height
    tmp = gph.copy()                  #Temperature
    vpr = gph.copy()                  #Vapor pressure
    if verbose:
        print('INFO: IMAGE DIMENSIONS: {} LATITUDES AND {} LONGITUDES'.format(nlat, nlon))

    lvls = 100.0*lvls              #Conversion to absolute pressure
    for i in range(nlvls):
        grbs.seek(gphind[i])   #Reading potential height.
        grb = grbs.read(3)
        gph[i,:,:] = grb[0].values[uu,:][:,vv]/g

        #Reading temperature
        temp = grb[1].values[uu,:][:,vv]
        tmp[i,:,:] = temp

        if humidity in ('R'):   # Relative humidity
            esat = cc_era(temp,cdic)       
            temp = grb[2].values[uu,:][:,vv]/100.0
            vpr[i,:,:] = temp*esat
            
        elif humidity in ('Q'):
            val = grb[2].values  #Specific humidity
            temp = grb[2].values[uu,:][:,vv]
            vpr[i,:,:] = temp*lvls[i]*alpha/(1+(alpha - 1)*temp)
            
        else:
            assert 1==0, 'Undefined Humidity in get_ecmwf().'     

    return lvls,latlist,lonlist,gph,tmp,vpr
###############Completed GET_ECMWF########################################



import numpy as np
import pygrib
import scipy.interpolate as si
import sys

#############Clausius-Clapeyron for NARR ###########
def cc_narr(tmp,cdic):
    '''Clausius Clayperon law used in the NARR model.

    Args:
        * tmp  (np.ndarray): Temperature.
        * cdic (dict)      : Dictionnary of constants.

    Returns:
        * esat (np.ndarray): Saturation water vapor partial pressure.'''

    a1w=cdic['a1w']
    a3w=cdic['a3w']
    a4w=cdic['a4w']
    T3=cdic['T3']
    Rv=cdic['Rv']
    esat = a1w*np.exp(a3w*(tmp-T3)/(tmp-a4w))

    return esat
###############Completed CC_NARR#####################################


########Read in ERA data from a given ERA Interim file##################
def get_narr(fname,minlat,maxlat,minlon,maxlon,cdic,verbose=False):
    '''Read data from NARR grib file. Note that the Lon values should be between [0-360].
    GRB file with weather model data can be downloaded from http://nomads.ncdc.noaa.gov/data/narr .

    Args:
        * fname  (str)  :  Path to the grib file
        * minlat (float):  Minimum latitude
        * maxlat (float):  Maximum latitude
        * minlon (float):  Minimum longitude
        * maxlon (float):  Maximum longitude
        * cdic   (float):  Dictionary of constants
    
    Kwargs:
        * humidity (str): Specific ('Q') or relative humidity ('R').

    Returns:
        * lvls    (np.ndarray): Pressure levels
        * latlist (np.ndarray): Latitudes of the stations
        * lonlist (np.ndarray): Longitudes of the stations
        * gph     (np.ndarray): Geopotential height
        * tmp     (np.ndarray): Temperature
        * vpr     (np.ndarray): Vapor pressure

    .. note::
        Uses cc_narr by default.
        '''

    if verbose:
        print('PROGRESS: READING GRIB FILE')
    lvls = np.array([100, 125, 150, 175, 200, 225, 250, 275, 300, 350, 400, 450, 500, 550,
                     600, 650, 700, 725, 750, 775, 800, 825, 850, 875, 900, 925, 950, 975, 1000])
    nlvls = len(lvls)

    alpha = cdic['Rv']/cdic['Rd']
    gphind = np.array([16,24,32,40,48,56,64,72,80,88,96,104,112,120,128,137,
                       146,155,164,173,182,191,200,210,219,228,237,246,255])
    
    grbs = pygrib.open(fname)
    grbs.seek(gphind[0])
    grb=grbs.read(1)[0]
    lats,lons = grb.latlons()
    lons[lons<0] += 360.
    g = cdic['g']    
    mask = (lats > minlat) & (lats < maxlat) & (lons > minlon) & (lons < maxlon)
    [ii,jj] = np.where(mask == True)
    del mask
    latlist = lats[ii,jj]
    lonlist = lons[ii,jj]
    nstn = len(ii)
    
    ####Create arrays for 3D storage
    gph = np.zeros((nlvls, nstn))     #Potential height
    tmp = gph.copy()                  #Temperature
    vpr = gph.copy()                  #Vapor pressure
    if verbose:
        print('Number of stations:', nstn)

    lvls = 100.0*lvls                 #Conversion to absolute pressure
    for i in range(nlvls):
        grbs.seek(gphind[i])          #Reading potential height.
        grb = grbs.read(3)
        val = grb[0].values
        gph[i,:] = val[ii,jj]

        val = grb[1].values           #Reading temperature
        temp = val[ii,jj]
        tmp[i,:] = temp

                
        val = grb[2].values  #Specific humidity
        temp = val[ii,jj]
        vpr[i,:] = temp*lvls[i]*alpha/(1+(alpha - 1)*temp)
        
    return lvls,latlist,lonlist,gph,tmp,vpr
###############Completed GET_ERA########################################


########Interpolates the NARR delay to a regular grid####################
def intdel(hgt,latlin,lonlin,delcin,spacing=0.3):
    '''Interpolates the NARR data to a regular grid with a grid spacing being the average of previous grid spaces

    Args:
            * hgt     (np.ndarray): Altitude levels
            * latlin  (np.ndarray): Latitudes of the stations
            * lonlin  (np.ndarray): Longitudes of the stations
            * delcin  (np.ndarray): Delay cube len(latlist)xlen(hgt)
    
    Returns:
            * latlist (np.ndarray): Latitudes of the stations
            * lonlist (np.ndarray): Longitudes of the stations
            * delc    (np.ndarray): Delay cube len(latlist)xlen(hgt)

    .. note::
    '''

    # Points array
    nstn = len(latlin)
    Points = np.zeros((nstn,2))
    Points[:,0] = lonlin
    Points[:,1] = latlin

    # Output arrays
    minlat = latlin.min()
    maxlat = latlin.max()
    minlon = lonlin.min()
    maxlon = lonlin.max()
    latlist = np.arange(minlat,maxlat,spacing)
    lonlist = np.arange(minlon,maxlon,spacing)
    [lonlist,latlist] = np.meshgrid(lonlist,latlist)
    latlist = latlist.flatten()
    lonlist = lonlist.flatten()

    # Xi arrays
    nstno = len(latlist)
    Xi = np.zeros((nstno,2))
    Xi[:,0] = lonlist
    Xi[:,1] = latlist

    # Create the arrays
    nlvls = len(hgt)
    delc = np.zeros((nstno,nlvls))

    # Loop on the pressure levels
    for n in range(nlvls):
        delc[:,n] = si.griddata(
            Points,
            delcin[:,n],
            Xi,
            method='cubic',
            #fill_value=10*float(n),
        )

    return delc,latlist,lonlist


############################################################
# Program is part of PyAPS                                 #
# Copyright 2012, by the California Institute of Technology#
# Contact: earthdef@gps.caltech.edu                        #
############################################################



############################################################
# Program is part of PyAPS                                 #
# Copyright 2012, by the California Institute of Technology#
# Contact: earthdef@gps.caltech.edu                        #
# Modified by A. Benoit and R. Jolivet 2019                #
# Ecole Normale Superieure, Paris                          #
# Contact: insar@geologie.ens.fr                           #
############################################################


import os.path
import sys
import numpy as np
import scipy.interpolate as si
from pyaps3 import utils, processor


##############Creating a class object for PyAPS use.
class PyAPS:
    '''Class for dealing with Atmospheric phase corrections in radar/geo coordinates.
    Operates on one weather model file and one Geo.rsc file at a time.'''

    def __init__(self, gribfile, dem, lat, lon, inc=0.0, mask=None,
                 grib='era5', humidity='Q', Del='comb', model='ERA5', verb=False):
        '''Initiates the data structure for atmos corrections in geocoded domain.
        Args:
            * gribfile (str)        : path to downloaded grib file
            * dem      (np.ndarray) : height    in size of (length, width)
            * lat      (np.ndarray) : latitude  in size of (length, width)
            * lon      (np.ndarray) : longitude in size of (length, width)

        Kwargs:
            * inc      (np.ndarray) : incidence angle (in size of (length, width) for np.ndarray)
            * mask     (np.ndarray) : mask of valid pixels in size of (length, width)
            * grib     (str)        : grib name in ['ERA5', 'ERAINT', 'HRES', 'NARR', 'MERRA']
            * humidity (str)        : ['Q', 'R']
            * Del      (str)        : ['comb', 'wet', 'dry']
            * model    (str)        : ECMWF dataset name in ['era5', 'eraint', 'hres']
            * verb     (bool)       : True or False

        .. note::
            For ISCE products, lat/lon can be read from lat/lon.rdr file
            For ROIPAC products, lat, lon = utils.get_lat_lon('radar_16rlks.hgt.rsc')
        '''

        #--------- Check files exist and we have what's needed for computation
        # check grib type and import module
        grib = grib.upper()
        if grib in ['ERA5','ERAINT','HRES']:
            from pyaps3 import era
        elif grib == 'NARR':
            from pyaps3 import narr
        elif grib == 'MERRA':
            from pyaps3 import merra
        else:
            raise ValueError('PyAPS: Undefined grib file source: {}'.format(grib))
        self.grib = grib

        # Check Humidity variable
        humidity = humidity.upper()
        assert humidity in ('Q','R'), 'PyAPS: Undefined humidity.'
        self.humidity = humidity
        if self.grib in ('NARR','MERRA'):
            assert self.humidity in ('Q'), 'PyAPS: Relative humidity not provided by NARR/MERRA.'

        # Check the model for ECMWF
        self.model = model

        # Check grib file exists
        assert os.path.isfile(gribfile), 'PyAPS: GRIB File does not exist.'
        self.gfile = gribfile

        # Get altitude, lon, lat, etc
        self.dem = dem
        self.lon = lon
        self.lat = lat
        self.inc = inc
        self.mask = np.ones(self.dem.shape) if mask is None else mask

        # Get size
        self.ny, self.nx = self.dem.shape
        assert self.lon.shape  == (self.ny, self.nx), 'PyAPS: Longitude array size mismatch'
        assert self.lat.shape  == (self.ny, self.nx), 'PyAPS: Latitude array size mismatch'
        assert self.mask.shape == (self.ny, self.nx), 'PyAPS: Mask array size mismatch'

        # check incidence angle size and type
        if isinstance(self.inc, np.ndarray):
            assert self.inc.shape == (self.ny, self.nx), 'PyAPS: Incidence array size mismatch'
            if verb:
                print('INFO: INCIDENCE ANGLE AS AN ARRAY')
        elif isinstance(self.inc, (int, float, np.float32, np.float64)):
            if verb:
                print('INFO: INCIDENCE ANGLE AS A NUMBER: {} DEG'.format(self.inc))
        else:
            raise ValueError('PyAPS: unrecognized incidence data type: {}'.format(type(self.inc)))

        #--------- Initialize variables
        self.dict = processor.initconst()

        # Get some scales
        if self.grib in ('ERA5','ERAINT','HRES'):
            self.hgtscale = ((self.dict['maxAlt'] - self.dict['minAlt']) / self.dict['nhgt']) / 0.703
            self.bufspc = 1.2
        elif self.grib in ('NARR'):
            self.hgtscale = ((self.dict['maxAlt'] - self.dict['minAlt']) / self.dict['nhgt']) / 0.3
            self.bufspc = 1.2
        elif self.grib in ('MERRA'):
            self.hgtscale = ((self.dict['maxAlt'] - self.dict['minAlt']) / self.dict['nhgt']) / 0.5
            self.bufspc = 1.0 

        # Problems in isce when lon and lat arrays have weird numbers
        if self.grib in ('ERA5','ERAINT','HRES'):
            self.lon[self.lon > 180.] -= 360.
        else:
            self.lon[self.lon < 0.] += 360.0
        self.minlon = np.nanmin(self.lon[np.nonzero(self.mask)]) - self.bufspc
        self.maxlon = np.nanmax(self.lon[np.nonzero(self.mask)]) + self.bufspc
        self.minlat = np.nanmin(self.lat[np.nonzero(self.mask)]) - self.bufspc
        self.maxlat = np.nanmax(self.lat[np.nonzero(self.mask)]) + self.bufspc
        if verb:
            print('INFO: AREA COVERAGE IN SNWE: ({:.2f}, {:.2f}, {:.2f}, {:.2f})'.format(
                self.maxlat, self.minlat, self.minlon, self.maxlon))

        #--------- Extract infos from gribfiles
        if self.grib in ('ERA'):
            assert False, 'Need to modify get_era to fit with the new standards'
            [lvls,latlist,lonlist,gph,tmp,vpr] = era.get_era(
                self.gfile,
                self.minlat,
                self.maxlat,
                self.minlon,
                self.maxlon,
                self.dict,
                humidity=self.humidity,
                verbose=verb,
            )

        elif self.grib in ('ERA5','ERAINT','HRES'):
            [lvls,latlist,lonlist,gph,tmp,vpr] = era.get_ecmwf(
                self.model,
                self.gfile,
                self.minlat,
                self.maxlat,
                self.minlon,
                self.maxlon,
                self.dict,
                humidity=self.humidity,
                verbose=verb,
            )

        elif self.grib in ('NARR'):
            assert False, 'Need to modify get_narr to fit with the new standards'
            [lvls,latlist,lonlist,gph,tmp,vpr] = narr.get_narr(
                self.gfile,
                self.minlat,
                self.maxlat,
                self.minlon,
                self.maxlon,
                self.dict,
                verbose=verb,
            )

        elif self.grib in ('MERRA'):
            assert False, 'Need to modify get_merra to fit with the new standards'
            [lvls,latlist,lonlist,gph,tmp,vpr] = merra.get_merra(
                self.gfile,
                self.minlat,
                self.maxlat,
                self.minlon,
                self.maxlon,
                self.dict,
                verbose=verb,
            )
            lonlist[lonlist < 0.] += 360.0

        # Make a height scale
        hgt = np.linspace(self.dict['minAltP'], gph.max().round(), self.dict['nhgt'])

        # Interpolate pressure, temperature and Humidity of hgt
        [Pi,Ti,Vi] = processor.intP2H(lvls, hgt, gph, tmp, vpr, self.dict, verbose=verb)

        # Calculate the delays
        [DDry,DWet] = processor.PTV2del(Pi,Ti,Vi,hgt,self.dict,verbose=verb)
        if Del.lower() == 'comb':
            Delfn = DDry+DWet
        elif Del.lower() == 'dry':
            Delfn = DDry
        elif Del.lower() == 'wet':
            Delfn = DWet
        else:
            raise ValueError('Unrecognized delay type: {}'.format(Del))

        if self.grib in ('NARR'):
            assert False, 'Need to check narr.intdel'
            [Delfn, latlist, lonlist] = narr.intdel(hgt, latlist, lonlist, Delfn)

        #--------- Save things
        self.Delfn = Delfn
        self.latlist = latlist
        self.lonlist = lonlist
        self.lat = lat
        self.lon = lon
        self.hgt = hgt
        self.Pi = Pi
        self.Ti = Ti
        self.Vi = Vi
        self.verb = verb

        # All done
        return


    def getdelay(self, dout=None, outFile=None, wvl=np.pi*4., writeStations=True):
        '''Get the 2D matrix of tropospheric delay with bilinear interpolation.
        Kwargs:
            * dout    : 2D np.ndarray, output delay matrix
            * outFile : str, file path of output delay matrix
            * wvl     : Wavelength in meters.
                        4*pi (by default) --> output results in delay in meters.
                        0.056 --> output results in delay in radians for C-band SAR.
        Returns:
            * dout    : 2D np.ndarray in size of (ny, nx) in float32.
        '''

        # To know the type of incidence (float, array)
        if isinstance(self.inc, (int, float, np.float32, np.float64)):
            cinc = np.cos(self.inc*np.pi/180.)
            incFileFlag = 'number'
        else:
            incFileFlag = 'array'

        # Get some info from the dictionary
        minAltp = self.dict['minAltP']

        # initiate output
        dout = np.zeros((self.ny, self.nx), dtype=np.float32) if dout is None else dout
        fout = open(outFile, 'wb') if outFile else None

        #######################################################################################
        # BILINEAR INTERPOLATION

        # Create the 1d interpolator to interpolate delays in altitude direction
        if self.verb:
            print('PROGRESS: FINE INTERPOLATION OF HEIGHT LEVELS')
        intp_1d = si.interp1d(self.hgt, self.Delfn, kind='cubic', axis=-1)

        # Interpolate the delay function every meter, for each station
        self.dem[np.isnan(self.dem)] = minAltp
        self.dem[self.dem < minAltp] = minAltp
        minH = np.max([np.nanmin(self.dem*self.mask), self.hgt.min()])
        maxH = int(np.nanmax(self.dem*self.mask)) + 100.
        kh = np.arange(minH,maxH)
        self.Delfn_1m = intp_1d(kh)
        self.alti = kh

        # no reshape
        Lonu = self.lonlist[0,:]
        Latu = self.latlist[:,0]

        # Create the cube interpolator for the bilinear method, to interpolate delays into a grid (x,y,z)
        if self.verb:
            print('PROGRESS: CREATE THE BILINEAR INTERPOLATION FUNCTION')

        # Define a linear interpolating function on the 3D grid: ((x, y, z), data)
        # We do the weird trick of [::-1,:,:] because Latu has to be in increasing order 
        # for the RegularGridInterpolator method of scipy.interpolate
        linearint = si.RegularGridInterpolator(
            (Latu[::-1], Lonu,kh),
            self.Delfn_1m[::-1,:,:],
            method='linear',
            bounds_error=False,
            fill_value=0.0,
        )

        # Show progress bar
        if self.verb:
            toto = utils.ProgressBar(maxValue=self.ny)
            print('PROGRESS: MAPPING THE DELAY')

        # Loop on the lines
        for m in range(self.ny):

            # Update progress bar
            if self.verb:
                toto.update(m+1, every=5)

            # Get latitude and longitude arrays
            lati = self.lat[m,:]*self.mask[m,:]
            loni = self.lon[m,:]*self.mask[m,:]

            # Remove negative values
            if self.grib in ('ERA5','ERAINT','HRES'):
                loni[loni > 180.] -= 360.
            else:
                loni[loni < 0.] += 360.

            # Remove NaN values
            ii = np.where(np.isnan(lati))
            jj = np.where(np.isnan(loni))
            xx = np.union1d(ii,jj)
            lati[xx]=0.0
            loni[xx]=0.0

            # Get incidence if file provided
            if incFileFlag == 'array':
                cinc = np.cos(self.mask[m,:]*self.inc[m,:]*np.pi/180.)

            # Make the bilinear interpolation
            D = self.dem[m,:]
            val = linearint(np.vstack((lati, loni, D)).T)*np.pi*4.0/(cinc*wvl)
            val[xx] = np.nan

            # save output
            dout[m,:] = val
            if outFile:
                val.astype(np.float32).tofile(fout)

        if self.verb:
            toto.close()

        if outFile:
            fout.close()

        return dout
